function celldata=evol_nb_cell(state,time_flag,save_flag)
% EVOL_NB_CELL(CELL_TYPE,TIME_FLAG,SAVE_FLAG)
%   Extract the number of cells of type CELL_TYPE (string) from ReSCAL logs
%   and plot the evolution.
%   If TIME_FLAG is set, read time data in t0 unit from TIME.log file.
%   If SAVE_FLAG=1, save figure.
%   If SAVE_FLAG=2, save data.
%   If SAVE_FLAG=3, save figure and data.

if ~exist('time_flag','var')
  time_flag=0;
end

if ~exist('save_flag','var')
  save_flag=0;
end

%% read header of CELL.log
fprintf('reading CELL.log\n');
fid=fopen('CELL.log');
if (fid==-1)
  fprintf('file not found\n');
  return
end
textscan(fid,'# CELL STATES');
C=textscan(fid,'NB_STATES = %d');
n=C{1};
CTypes=textscan(fid,'ST(%d): %s %s',n);
id_cell=-1;
for i=1:n
  if strcmp(CTypes{2}(i),state)
    celltype=CTypes{1}(i);
    id_cell=celltype+1;
    fprintf('id of state %s: %d\n', state, celltype);
  end
end

if (id_cell < 0)
  fprintf('state %s not found\n', state);
  return
end

%% skip line of text
C=textscan(fid,'%[^\n]s');
%celldisp(C)

%% read cell data
C=textscan(fid,'%d','delimiter',':');
data0=cell2mat(C);
l=size(data0,1);
nb=l/(n+1);
fprintf('number of records: %d\n', nb);
data=reshape(data0,n+1,nb)';
celldata=data(:,id_cell+1);
fclose(fid);

%% read time data
if time_flag>0
  fprintf('reading TIME.log\n');
  fid=fopen('TIME.log');
  if (fid==-1)
    fprintf('file not found\n');
    return
  end
  C=textscan(fid,'%[^\n]s');
  %celldisp(C)
  C=textscan(fid,'%f','delimiter',':dhms');
  data0=cell2mat(C);
  l=size(data0,1);
  c=l/nb;
  data=reshape(data0,c,nb)';
  timedata=data(:,4);
  fclose(fid);
end

%% plot data
lw=1;
figure
hold on
if time_flag==0
  plot(celldata,'linewidth',lw)
  xlabel('Computation time (rec. number)','interpreter','tex')
else
  plot(timedata,celldata,'linewidth',lw)
  xlabel('Time (t_0)','interpreter','tex')
end
title(sprintf('Evolution of the number of %s cells', state),'interpreter','tex')
ylabel('Number of cells','interpreter','tex')
grid on

avg=mean(celldata);
fprintf('average number of %s cells: %f\n', state, avg);

if save_flag
  if bitand(save_flag,1)
    figname=sprintf('EVOL_NB_CELL_%s.png',state);
    fprintf('save figure into file: %s\n',figname);
    print(figname,'-dpng','-r250');
  end
  if bitand(save_flag,2)
    filename=sprintf('EVOL_NB_CELL_%s.out',state);
    fprintf('save data into file: %s\n',filename);
    if time_flag==0
      data=double(celldata);
    else
      data=[timedata double(celldata)];
    end
    save(filename,'data','-ascii');
  end
end

return

