/* ReSCAL - Csp templates
 *
 * Copyright (C) 2022
 *
 * Author: Olivier Rozier <rozier@ipgp.fr>
 *
 * This file is part of ReSCAL.
 *
 * ReSCAL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "defs.h"
#include "macros.h"
#include "cells.h"
#include "templates.h"
#include "space.h"
#include "param.h"

extern int prog;    //main prog
extern int use_genesis;
extern int boundary;  //boundary conditions
extern int  H, L, D, HL, HLD;   // size of cellular space
extern int L_border, H_border, D_border;  // borders
extern Cell *TE;          // cellular space

extern float csp_angle;
extern char *rot_map;
extern int rot_mode;

char *csp_template_str = NULL; //value of Csp_template parameter
char *csp_template_add_str = NULL; //value of Csp_template_add parameter
CSP_Template csp_templates[MAX_TEMPLATES]; //array of CSP templates of the current simulation
int nb_templates=0;
int nb_templates_add=0;
CSP_Template t_templates[MAX_TEMPLATES]; //available templates
int nb_template_types = 0; //number of template types
int flag_restore_template = 0;


double drand48();
void  srand48();


void params_templates()
{
  param_family("TEMPLATE", "Template parameters");
  parameter("Csp_template", "template array for the generation of the initial cellular space (syntax: temp1 + temp2 + ...)", &csp_template_str, PARAM_STRING,"TEMPLATE");
  parameter("Add_csp_template", "add one or more templates to a cellular space from file (syntax: temp1 + temp2 + ...)", &csp_template_add_str, PARAM_STRING,"TEMPLATE");
  if (prog == PROG_RESCAL) parameter("Restore_template", "restore template after rotation in case of static cells (YES|NO) - NO by default", &flag_restore_template, PARAM_BOOLEAN, "TEMPLATE");
}


void init_temp(int type, char *name, char *desc, int nb_args, ...)
{
  va_list vl;
  int i;

  t_templates[nb_template_types].type = type;
  t_templates[nb_template_types].name = name;
  t_templates[nb_template_types].desc = desc;
  t_templates[nb_template_types].nb_args = nb_args;

  AllocMemory(t_templates[nb_template_types].args, float, nb_args);
  va_start(vl, nb_args);
  for (i=0; i<nb_args; i++){
    t_templates[nb_template_types].args[i] = (float) va_arg(vl, double);
    //LogPrintf("t_templates[%d].args[%d] = %f\n", nb_template_types, i, t_templates[nb_template_types].args[i]);
  }
  va_end(vl);
  nb_template_types++;
}

/// TEMPLATES USER'S GUIDE
/// ----------------------
///
/// A CSP template is a predefined cellular space configuration, with optional arguments.
/// The user gives to the parameter Csp_template a value of the form:
/// TEMPLATE_NAME, or TEMPLATE_NAME(arg1, ...).
///
/// Example: LAYER(5) generates a sand layer of 5 cells height (in DUN model).
///
/// One may now combine several templates separated by a '+' character. Note that the order matters when various templates overlap each other.
/// The priority strictly increases from the first to the last template in the list.
///
/// Example: LAYER(5) + CONE generates a cone of sand surrounded by a uniform layer of sand (in DUN model).
///
/// Here we briefly explain how to create a new CSP template:
/// 1) Add a new type in CSP_TEMPLATES (in templates.h).
/// 2) Define the new template in init_template_types() by adding a line of the form
///    init_template(template type, "TEMPLATE NAME", "Template usage description", number of optional arguments, default value for argument 1, default value for argument 2, ...);
///    Note: Default values of arguments MUST be floats (in case of an integer, just add a zero fractional part).
/// 3) In the function init_cell_from_template(), write the implementation for the new CSP template.
///


/// Description of available templates
void init_template_types()
{
  if (prog!=PROG_RESCAL) init_temp(CSP_CUSTOM, "CUSTOM", "CUSTOM:\t\t\tcustom template", 0);
  init_temp(CSP_BLOCK, "BLOCK", "BLOCK(xmin, xmax, ymin, ymax, zmin, zmax, type=0):\t3D block of <type> cells", 7, 0.0, 10.0, 0.0, 10.0, 0.0, 10.0, 0.0);
#ifdef MODEL_DUN
  init_temp(CSP_LAYER, "LAYER", "LAYER(h=1.0):\t\t\tsand layer of height <h>", 1, 1.0);
#ifdef CELL_COLOR
  init_temp(CSP_LAYER_COL, "LAYER_COL", "LAYER_COL(h=1.0, d=0.5):\tbinary sand layer (two colors of grain) of height <h> and density <d> for white color", 2, 1.0, 0.5);
#endif
  init_temp(CSP_HBLOCK, "HBLOCK", "HBLOCK(xmin, xmax, ymin, ymax, h, type=0):\t3D block of <type> cells and height <h> resting on the ground", 6, 30.0, 50.0, 50.0, 150.0, 20.0, (double)GR);
  init_temp(CSP_CYLINDER, "CYLINDER", "CYLINDER(h=20, w=100):\tcylinder of height <h> and width <w>", 2, 20.0, 100.0);
  init_temp(CSP_CONE, "CONE", "CONE(h=30, w=100, x=L/2, y=D/2, type):\t\tcone of height <h> and width <w> centered on (x,y)", 5, 30.0, 100.0, 0.0, 0.0, (double)GR);
  init_temp(CSP_RCONE, "RCONE", "RCONE(rh=0.66, rw=0.4, rx=0.5, ry=0.5, type):\tcone of relative height <rh>*H, relative width <rw>*Min(L,D) and centered on relative position (<rx>*L,<ry>*D)", 5, 0.66, 0.4, 0.5, 0.5, (double)GR);
#ifdef CELL_COLOR
  init_temp(CSP_CONE2, "CONE2", "CONE2:\t\t2 cones with different colors and sizes", 0);
#endif
  init_temp(CSP_RWALL, "RWALL", "RWALL(rh=0.5, rx=0.3):\twall of relative height <rh>*H and relative position <rx>*L", 2, 0.5, 0.3);
  init_temp(CSP_WAVES_2D, "WAVES_2D", "WAVES_2D(per_min=10, per_max=60, ufd=0, amp=4, mh=30):\t2d sand waves with increasing periods from <per_min> to <per_max>, amplitude <amp> and mean height <mh> for stability analysis; it has uniform frequency distribution when <ufd> flag is set", 5, 10.0, 60.0, 0.0, 4.0, 30.0);
  init_temp(CSP_WAVY_NS_LAYER, "WAVY_NS_LAYER", "WAVY_NS_LAYER(per=40, amp=15, h=5):\tnorth-south sand layer with wavy east boundary of period <per>, amplitude <amp> and height <h> for stability analysis", 3, 40.0, 15.0, 5.0);
  init_temp(CSP_TRIANGLES, "TRIANGLES", "TRIANGLES(n=8, ah=10, mh=20):\t<n> east-west periodic triangles of vertical amplitude <ah> and mean height <mh>", 3, 8.0, 10.0, 20.0);
  init_temp(CSP_SRC_CEIL, "SRC_CEIL", "SRC_CEIL:\t\t\tsource of sand covering the ceiling", 0);
  init_temp(CSP_SRC_UPWIND_CEIL, "SRC_UPWIND_CEIL", "SRC_UPWIND_CEIL:\t\t\tsource of sand in the ceiling parallel to the upwind boundaries", 0);
  init_temp(CSP_SRC_DISK, "SRC_DISK", "SRC_DISK(w=40, x=100, y=D/2):\tcircular source of sand on the ground with width <w> and centered on (x,y)", 3, 40.0, 100.0, 0.0);
  init_temp(CSP_SRC_DISK_CEIL, "SRC_DISK_CEIL", "SRC_DISK_CEIL(w=10, x=100, y=D/2):\tcircular source of sand in the ceiling with width <w> and centered on (x,y)", 3, 10.0, 100.0, 0.0);
  init_temp(CSP_SRC_RECT, "SRC_RECT", "SRC_RECT(lx=5, ly=D, x=100, y=D/2):\trectangular source of sand on the ground with dimensions (lx,ly) and centered on (x,y)", 4, 5.0, 0.0, 100.0, 0.0);
  init_temp(CSP_SRC_RECT_CEIL, "SRC_RECT_CEIL", "SRC_RECT_CEIL(lx=5, ly=D, x=100, y=D/2):\trectangular source of sand in the ceiling with dimensions (lx,ly) and centered on (x,y)", 4, 5.0, 0.0, 100.0, 0.0);
  init_temp(CSP_OUT_EAST, "OUT_EAST", "OUT_EAST(w=5):\t\tsink of sand of width <w> along the east border", 1, 5.0);
  init_temp(CSP_OUT_NORTH, "OUT_NORTH", "OUT_NORTH(w=5):\t\tsink of sand of width <w> along the north border", 1, 5.0);
  init_temp(CSP_OUT_SOUTH, "OUT_SOUTH", "OUT_SOUTH(w=5):\t\tsink of sand of width <w> along the south border", 1, 5.0);
  init_temp(CSP_STICKS, "STICKS", "STICKS(h=10, w=50, d=0.1, x=100, y=D/2):\tcircular area of width <w> centered on (x,y), with vertical sticks of height <h> and density <d>", 5, 10.0, 50.0, 0.1, 100.0, 0.0);
  init_temp(CSP_SMILEY, "SMILEY", "SMILEY:\t\t\t:-)", 0);
#elif defined(MODEL_AVA)
  init_temp(CSP_SRC_CEIL, "SRC_CEIL", "SRC_CEIL:\t\t\tsource of sand covering the ceiling", 0);
  init_temp(CSP_SRC_DISK_CEIL, "SRC_DISK_CEIL", "SRC_DISK_CEIL(w=10, x=L/2, y=D/2):\tcircular source of sand in the ceiling with width <w> and centered on (x,y)", 3, 10.0, 0.0, 0.0);
#elif defined(MODEL_CMB)
  init_temp(CSP_LAYER, "LAYER", "LAYER(h=H/2):\t\tfluid layer of height <h>", 1, 0.0);
#elif defined(MODEL_ICB)
  init_temp(CSP_LAYER, "LAYER", "LAYER(h=H*2/3):\t\tfluid layer of height <h> on top", 1, 0.0);
#elif defined(MODEL_RIV) || defined(MODEL_D2G)
  init_temp(CSP_LAYER, "LAYER", "LAYER(h=H/2):\t\tsolid layer of height <h>", 1, 0.0);
  init_temp(CSP_SLOPE, "SLOPE", "SLOPE(r=1.0,h=0.0):\t\tsloping ground of gradient <r> starting at height <h>", 2, 1.0, 0.0);
  init_temp(CSP_CLOSED_BORDER, "CLOSED_BORDER", "CLOSED_BORDER\t\tclosed boundaries", 0);
  init_temp(CSP_OPEN_BORDER, "OPEN_BORDER", "OPEN_BORDER(h=1.0)\t\topen boundaries above a given height <h>", 1, 1.0);
  init_temp(CSP_OPEN_BORDER1, "OPEN_BORDER1", "OPEN_BORDER1(h=1.0,w=L)\t\tclosed boundaries except a rectangular outlet at height <h> and of width <w> on the south border", 2, 1.0, 0.0);
  init_temp(CSP_SRC, "SRC", "SRC(x=L/2, y=1):\tsource in the ceiling at (x,y)", 2, 0.0, 1.0);
#elif defined(MODEL_LIFE)
  init_temp(CSP_RANDOM, "RANDOM", "RANDOM(p=0.5):\t\tcells have probability <p> to be alive", 1, 0.5);
#endif
  //LogPrintf("init_template_types: %d templates\n", nb_template_types);
}


int parse_template(int idx, char *str_temp, int flag_add)
{
  char *ptr;
  int i;
  int err=0;
  CSP_Template *p_temp = csp_templates + idx;

  LogPrintf("template index: %d%s\n", idx, flag_add?" (add)":"");

  if (idx >= MAX_TEMPLATES){
    ErrPrintf("ERROR: too many templates (%d)\n", nb_templates);
    exit(-1);
  }

  //read template name
  ptr = strtok(str_temp,"(");
  LogPrintf("template name: %s\n", ptr);

  //find and copy the full template
  for (i=0; (i < nb_template_types) && strcmp(ptr,t_templates[i].name); i++);
  if (i < nb_template_types){
    memcpy(p_temp, &t_templates[i], sizeof(CSP_Template));
    p_temp->flag_add = flag_add;
    LogPrintf("template type: %d\n", p_temp->type);
    AllocMemory(p_temp->args, float, p_temp->nb_args);
    memcpy(p_temp->args, t_templates[i].args, p_temp->nb_args*sizeof(float));
  }
  else{
    if (!strcmp(ptr,"CUSTOM")){
      if ((prog == PROG_RESCAL) && !use_genesis){
        WarnPrintf("WARNING: CUSTOM template not supported in rescal, please use genesis first and set \"Use_genesis = YES\" in the parameter file\n");
      }
      else
        return -1;
    }
    ErrPrintf("ERROR: Incorrect template value, %s\n", str_temp);
    exit(-2);
  }

  //read template optional arguments, if any
  ptr = strtok(NULL,",)");
  i = 0;
  while (ptr && (i < p_temp->nb_args)){
    //LogPrintf("csp_template.args[%d] = %s\n", i, ptr);
    p_temp->args[i++] = read_float(ptr, &err);
    if (err){
      ErrPrintf("ERROR: bad argument \"%s\" for template %d (%s)\n", ptr, idx, p_temp->name);
      exit(-1);
    }
    ptr = strtok(NULL,",)");
  }
  if (ptr){
    ErrPrintf("ERROR: Too many arguments in Csp_template (%s)\n", ptr);
    exit(-2);
  }
  for (i=0; i < p_temp->nb_args; i++){
    LogPrintf("csp_templates[%d].args[%d] = %f\n", idx, i, p_temp->args[i]);
  }

  nb_templates++;

  return i;
}

/// Check if template type is present in the current array of templates
int search_template(int type)
{
  int id_temp;
  for (id_temp=0; id_temp<nb_templates; id_temp++){
    if (csp_templates[id_temp].type == type) break;
  }

  if (id_temp==nb_templates) id_temp=-1;
  //LogPrintf("search_template: type=%d, id_temp=%d\n",type,id_temp);

  return id_temp;
}

void parse_template_array()
{
  int idx=0;
  int ret;

  nb_templates = 0;
  nb_templates_add = 0;

  if (csp_template_str){
    /// split string
    char *ptr = strsep(&csp_template_str,"+");
    //LogPrintf("ptr=%s\n", ptr);
    //LogPrintf("csp_template_str=%s\n", csp_template_str);
    while (ptr && (idx<MAX_TEMPLATES)){
      ret = parse_template(idx, ptr, 0);
      //LogPrintf("ret=%d\n", ret);
      //if (ret>=0) nb_templates++;
      ptr = strsep(&csp_template_str,"+");
      //LogPrintf("ptr=%s\n", ptr);
      //LogPrintf("csp_template_str=%s\n", csp_template_str);
      idx++;
    }
  }

  if (csp_template_add_str){
    /// split string
    char *ptr = strsep(&csp_template_add_str,"+");
    //LogPrintf("ptr=%s\n", ptr);
    //LogPrintf("csp_template_add_str=%s\n", csp_template_add_str);
    while (ptr && (idx<MAX_TEMPLATES)){
      parse_template(idx, ptr, 1);
      ptr = strsep(&csp_template_add_str,"+");
      //LogPrintf("ptr=%s\n", ptr);
      //LogPrintf("csp_template_add_str=%s\n", csp_template_add_str);
      nb_templates_add++;
      idx++;
    }
    nb_templates = idx;
  }

#ifdef MODEL_DUN
  if ((boundary == BC_REINJECTION) && (search_template(CSP_SRC_UPWIND_CEIL)==-1)){
    LogPrintf("add template SRC_UPWIND_CEIL for reinjection\n");
    parse_template(idx++, "SRC_UPWIND_CEIL", 1);
    nb_templates_add++;
    nb_templates++;
    if ((rot_mode == ROT_MODE_FULL) && !flag_restore_template){
      WarnPrintf("WARNING: please set Restore_template parameter to rebuild the source upwind as needed for reinjection\n");
    }
  }
#endif // MODEL_DUN

  if ((nb_templates == 0)){
    /// CSP_CUSTOM by default
    csp_templates[0].name = "CUSTOM";
    csp_templates[0].type = 0;
    csp_templates[0].nb_args = 0;
    nb_templates++;
    LogPrintf("CUSTOM template by default\n");
  }

  LogPrintf("nb_templates = %d\n", nb_templates);
}

///////////////////////////////////////
/// Set default values of template parameters that depend on other parameters (like H, L, D)
/// and make some verifications of integrity.
/// Note: The dimensions H, L, D do not integrate the bordures at this stage.
///////////////////////////////////////
void init_template_array()
{
  int id_temp, i;
  CSP_Template *p_temp = csp_templates;

  int Ld2 = L/2;
  int Dd2 = D/2;
  float Ldx= Ld2-0.5;
  float Ldy= Dd2-0.5;

  for (id_temp=0; id_temp<nb_templates; id_temp++, p_temp++){
    switch (p_temp->type){
#ifdef MODEL_DUN
      case CSP_LAYER:
      {
        /// format: LAYER(h)
        if (p_temp->args[0]<0) {ErrPrintf("ERROR: bad parameter in LAYER template, <h> must be >= 0 (%f)\n", p_temp->args[0]); exit(-1);}
        break;
      }

      case CSP_LAYER_COL:
      {
        /// format: LAYER_COL(h,d)
        if (p_temp->args[0]<0) {ErrPrintf("ERROR: bad parameter in LAYER_COL template, <h> must be >= 0 (%f)\n", p_temp->args[0]); exit(-1);}
        if (p_temp->args[1]>1){ErrPrintf("ERROR: bad parameter in LAYER_COL template, <d> must be <= 1 (%f)\n", p_temp->args[1]); exit(-1);}
        break;
      }

      case CSP_CONE:
      {
        /// format: CONE(h=30, w=100, x=L/2, y=D/2, type)
        if (!p_temp->args[2]) p_temp->args[2] = Ldx;
        if (!p_temp->args[3]) p_temp->args[3] = Ldy;
        break;
      }

      case CSP_WAVES_2D:
      {
        /// format: WAVES_2D(per_min=10, per_max=60, ufd=0, amp=4, mh=30)
        float per_min = p_temp->args[0];
        if (per_min<=0){ErrPrintf("ERROR: bad parameter in WAVES_2D template, <per_min> must be positive (%f)\n", per_min); exit(-1);}
        break;
      }

      case CSP_SRC_DISK:
      case CSP_SRC_DISK_CEIL:
      {
        /// format: SRC_DISK(w=40, x=100, y=D/2)
        if (!p_temp->args[2]) p_temp->args[2] = D/2.0;
        break;
      }

      case CSP_SRC_RECT:
      case CSP_SRC_RECT_CEIL:
      {
        /// format: SRC_RECT(lx=5, ly=D, x=100, y=D/2)
        if (!p_temp->args[1]) p_temp->args[1] = D;
        if (!p_temp->args[3]) p_temp->args[3] = D*0.5;
        break;
      }

#endif // MODEL_DUN

#ifdef MODEL_AVA
      case CSP_SRC_DISK_CEIL:
      {
        /// format: SRC_DISK_CEIL(w=10, x=L/2, y=D/2)
        if (!p_temp->args[1]) p_temp->args[1] = L*0.5;
        if (!p_temp->args[2]) p_temp->args[2] = D*0.5;
        break;
      }
#endif // MODEL_AVA
    }
    /// display parameters
    LogPrintf("checked template: %s(",p_temp->name);
    for (i=0; i<p_temp->nb_args; i++) LogPrintf("%s%f", (i? "," : ""), p_temp->args[i]);
    LogPrintf(")\n");
  }

}

void init_templates()
{
  init_template_types();
  parse_template_array();
  init_template_array();
}

///////////////////////////////////////
/// Default state of a cell
///////////////////////////////////////
void init_cell_default(Cell *aux)
{
#if defined(MODEL_DUN) || defined(MODEL_AVA) || defined(MODEL_CRY) || defined(MODEL_D2G)
  aux->celltype = AIR;  /// fluid
#elif defined(MODEL_CMB) || defined(MODEL_ICB)
  aux->celltype = PLUS;  /// solid
#elif defined(MODEL_RIV)
  aux->celltype = EAU;  /// fluid (~rain)
#elif defined(MODEL_DIF)
  aux->celltype = ZERO;  /// gas
#elif defined(MODEL_LIFE)
  aux->celltype = DEAD;
#endif
}

///////////////////////////////////////
/// Apply one template to one cell
///////////////////////////////////////
void init_cell_from_template(Cell *aux, int i, int j, int k, int id_temp)
{
  CSP_Template *p_temp = csp_templates + id_temp;

  /// use dimensions with no margin and no border
  int L0 = L-2*L_border;
  int D0 = D-2*D_border;
  int H0 = H-2*H_border;
  int Ld2 = L0/2;
  int Dd2 = D0/2;
  int Hd2 = H0/2;
  float Ldx= Ld2-0.5;
  float Ldy= Dd2-0.5;

  //if (!i && !j && !k) LogPrintf("init_cell_from_template - id_temp = %d, template type %d\n", id_temp, p_temp->type); //debug

  switch (p_temp->type){
/****************************************************************************/
/**************************** Generic templates *****************************/
/****************************************************************************/
    case CSP_BLOCK:
    {
      /// CSP_BLOCK: 3-dimensionnal block
      /// format: BLOCK(xmin, xmax, ymin, ymax, zmin, zmax, type)
      int xmin = (int)p_temp->args[0];
      int xmax = (int)p_temp->args[1];
      int ymin = (int)p_temp->args[2];
      int ymax = (int)p_temp->args[3];
      int zmin = (int)p_temp->args[4];
      int zmax = (int)p_temp->args[5];
      int type = (int)p_temp->args[6];
      int z = H0-1-j; /// distance from the bottom
      if ((i >= xmin) && (i < xmax) && (k >= ymin) && (k < ymax) && (z >= zmin) && (z < zmax)){
        aux->celltype = type;
      }
      break;
    }

/****************************************************************************/
/************************** Templates of DUN model **************************/
/****************************************************************************/
#ifdef MODEL_DUN
    case CSP_LAYER:
    {
      /// CSP_LAYER: sand layer
      /// format: LAYER(h)
      int hh = (int)p_temp->args[0];
      float frac = fmodf(p_temp->args[0],1.0);
      if ((j >= H0-1-hh) && (j<H0-1)) aux->celltype = GR;
      if ((j == H0-2-hh) && frac){ //h is not an integer
        if (drand48() < frac) aux->celltype = GR;
      }
      break;
    }

#ifdef CELL_COLOR
    case CSP_LAYER_COL:
    {
      /// CSP_LAYER_COL: binary sand layer
      /// format: LAYER_COL(h,d)
      int hh = (int)p_temp->args[0];
      float density = p_temp->args[1];
      float frac = fmodf(p_temp->args[0],1.0);
      if ((j >= H0-1-hh) && (j<H0-1)) aux->celltype = GR;
      if ((j == H0-2-hh) && frac){ //h is not an integer
        if (drand48() < frac) aux->celltype = GR;
      }
      if (aux->celltype == GR) aux->color = (drand48()<density) ? 1 : 0;
      break;
    }
#endif

    case CSP_HBLOCK:
    {
      /// CSP_HBLOCK: 3-dimensionnal block on the ground
      /// format: HBLOCK(xmin, xmax, ymin, ymax, h, type)
      int xmin = (int)p_temp->args[0];
      int xmax = (int)p_temp->args[1];
      int ymin = (int)p_temp->args[2];
      int ymax = (int)p_temp->args[3];
      int hh = (int)p_temp->args[4];
      int type = (int)p_temp->args[5];
      if ((i >= xmin) && (i < xmax) && (k >= ymin) && (k < ymax) && (j<=H0-2) && (j > H0-2-hh)){
        aux->celltype = type;
      }
      break;
    }

    case CSP_CYLINDER:
    {
      /// CSP_CYLINDER: sand cylinder
      /// format: CYLINDER(h, w)
      int h = p_temp->args[0];
      int w = p_temp->args[1];
      float di = i - Ldx;
      float dk = k - Ldy;
      if ((j >= (H0-h-1)) && (di*di + dk*dk < 0.25*w*w)) aux->celltype = GR;
      break;
    }

    case CSP_CONE:
    {
      /// CSP_CONE: sand cone
      /// format: CONE(h, w, x, y, type)
      float hcone = p_temp->args[0];
      float lcone = p_temp->args[1];
      float x = Round(p_temp->args[2] - 0.5);
      float y = Round(p_temp->args[3] - 0.5);
      int type = (int)p_temp->args[4];
      if (!hcone) break;
      float rc = 0.5*lcone*(hcone-((H0-2)-j))/hcone;
      if ((j<H0-1) && (j > (H0-2-hcone)) && ((i-x)*(i-x) + (k-y)*(k-y) <= rc*rc)) aux->celltype = type;
      break;
    }

    case CSP_RCONE:
    {
      /// CSP_RCONE: sand cone with relative dimensions and position
      /// format: RCONE(rh, rw, rx, ry, type)
      float hcone = p_temp->args[0]*(H0-2);
      float lcone = p_temp->args[1]*Min(L0,D0);
      float x = Round(p_temp->args[2]*L0 - 0.5);
      float y = Round(p_temp->args[3]*D0 - 0.5);
      int type = (int)p_temp->args[4];
      if (!hcone) break;
      float rc = 0.5*lcone*(hcone-((H0-2)-j))/hcone;
      if ((j<H0-1) && (j > (H0-2-hcone)) && ((i-x)*(i-x) + (k-y)*(k-y) <= rc*rc)) aux->celltype = type;
      break;
    }

    case CSP_CONE2:
    {
      /// CSP_CONE2: 2 sand cones for dune collisions, with different colors
      /// format: CONE2
      int hcone = H0/2;
      float lcone0 = D0/2;
      float lcone1 = D0/3;
      float rc0 = 0.5*lcone0*(j-(H0-hcone))/(hcone-1);
      float rc1 = 0.5*lcone1*(j-(H0-hcone))/(hcone-1);
      float di0 = i - L0/3;
      float dk0 = k - D0*3/5;
      float di1 = i - L0/8;
      float dk1 = k - D0/3;
      if ((j > (H0-hcone)) && (di0*di0 + dk0*dk0 < rc0*rc0)) aux->celltype = GR;
      if ((j > (H0-hcone)) && (di1*di1 + dk1*dk1 < rc1*rc1)){
        aux->celltype = GR;
  #ifdef CELL_COLOR
        aux->color = 1;
  #endif
      }
      break;
    }

    case CSP_RWALL:
    {
      /// CSP_RWALL(rh,rx): north-south wall or DUM cells
      int h = p_temp->args[0]*H0;
      int x = p_temp->args[1]*L0;
      if ((i>=x-1) && (i<=x+1) && (j>=H0-h)) aux->celltype = DUM;
      break;
    }

    case CSP_WAVES_2D:
    {
      /// CSP_WAVES_2D: 2d sand waves with increasing periods for stability analysis
      /// format: WAVES_2D(per_min=10, per_max=60, ufd=0, amp=4, mh=30)
      float periode;
      float per_min = p_temp->args[0];
      float per_max = p_temp->args[1];
      float ufd_flag = p_temp->args[2];
      float amp = p_temp->args[3];
      int mh = (int)p_temp->args[4];
      if (ufd_flag){
        // uniform frequency distribution
        float fmin = 1.0/per_max;
        float fmax = 1.0/per_min;
        float f = fmax + (fmin-fmax)*k/D0;
        periode = 1.0/f;
      }
      else
        periode = per_min + (per_max-per_min)*k/D0;
      if ((i==0) && (j==0) && (k%2==0)) LogPrintf("k = %d, periode = %f\n", k, periode);
      if (j > (H0-mh)-0.5*amp*sin(i*2*PI/periode)) aux->celltype = GR; //variable ondulations
      if ((k & 1) || (j > H0) || (j==0)) aux->celltype = DUM; //ground + ceiling + walls east-west
      break;
    }

    case CSP_WAVY_NS_LAYER:
    {
      /// CSP_WAVY_NS_LAYER: north-south sand layer with wavy east boundary for stability analysis
      float per = p_temp->args[0];
      float amp = p_temp->args[1];
      int hh = (int)p_temp->args[2];
      int i_wave = (int)(40 + 2 + 0.5*amp*(1+sin(k*2*PI/per)));
      if ((i >= 40) && (i <= i_wave) && (j <= H0) && (j > H0-hh)) aux->celltype = GR; //rangee de sable avec front de forme sinusoidale
      break;
    }

    case CSP_TRIANGLES:
    {
      /// CSP_TRIANGLES: east-west periodic triangles
      float lx0 = 0.5*L0/p_temp->args[0];
      float lx1 = fmodf(i, lx0*2);
      int hh=p_temp->args[1];
      int mh=p_temp->args[2];
      if (((lx1<lx0) && (j>H0-2-mh+hh/2-(hh*lx1/lx0))) || ((lx1>=lx0) && (j>H0-2-mh+hh/2-(hh*(2*lx0-lx1)/lx0)))) aux->celltype = GR;
      break;
    }

    case CSP_SRC_CEIL:
    {
      /// CSP_SRC_CEIL: source of sand covering the ceiling
      if (j == 0) aux->celltype = IN;
      break;
    }

    case CSP_SRC_UPWIND_CEIL:
    {
      /// CSP_SRC_UPWIND_CEIL: source of sand in the ceiling parallel to the upwind boundaries
      if ((j==0) && (i==5)) aux->celltype = IN;
      break;
    }

    case CSP_SRC_DISK:
    case CSP_SRC_DISK_CEIL:
    {
      /// CSP_SRC_DISK(w,x,y): circular sand source of width <w>
      float w = p_temp->args[0];
      float x = Round(p_temp->args[1] - 0.5);
      float y = Round(p_temp->args[2] - 0.5);
      float z = (p_temp->type == CSP_SRC_DISK) ? H0-1 : 0;
      if ((j == z) && ((i-x)*(i-x) + (k-y)*(k-y) <= w*w/4.0)) aux->celltype = IN;
      break;
    }

    case CSP_SRC_RECT:
    case CSP_SRC_RECT_CEIL:
    {
      /// CSP_SRC_RECT(lx,ly,x,y): rectangular source of sand
      float dx = p_temp->args[0]*0.5;
      float dy = p_temp->args[1]*0.5;
      float x = p_temp->args[2];
      float y = p_temp->args[3];
      float z = (p_temp->type == CSP_SRC_RECT) ? H0-1 : 0;
      if ((j == z) && (i >= x-dx) && (i < x+dx) && (k >= y-dy) && (k < y+dy)) aux->celltype = IN;
      break;
    }

    case CSP_OUT_EAST:
    {
      /// CSP_OUT_EAST(w): sink of sand of width <w> along the east border
      float w = p_temp->args[0];
      if ((j == H0-1) && (i >= L0-w)) aux->celltype = OUT;
      break;
    }

    case CSP_OUT_NORTH:
    {
      /// CSP_OUT_NORTH(w): sink of sand of width <w> along the north border
      float w = p_temp->args[0];
      if ((j == H0-1) && (k <= w-1)) aux->celltype = OUT;
      break;
    }

    case CSP_OUT_SOUTH:
    {
      /// CSP_OUT_SOUTH(w): sink of sand of width <w> along the south border
      float w = p_temp->args[0];
      if ((j == H0-1) && (k >= D0-w)) aux->celltype = OUT;
      break;
    }

    case CSP_STICKS:
    {
      /// CSP_STICKS: circular area of width <w> centered on (x,y), with vertical sticks of height <h> and density <d>
      int h = (int)p_temp->args[0];
      float w = p_temp->args[1];
      float density = p_temp->args[2];
      float x = p_temp->args[3];
      float y = p_temp->args[4];
      if (y == 0) y = D0*0.5;
      if ((j <= H0) && (j > H0-h) && ((i-x)*(i-x) + (k-y)*(k-y) <= w*w/4.0)) aux->celltype = ( j == H0-h+1 )? ( drand48()<density )? DUM : AIR : (aux-L)->celltype;
      break;
    }

    case CSP_SMILEY:
      /// CSP_SMILEY: :-)
      if (j>H0/2 && ((cos(PI*(i-L0/2)/L0/2)*cos(PI*(k-D0/2)/D0/2)<0.8 && cos(PI*(i-L0/2)/L0/2)*cos(PI*(k-D0/2)/D0/2)>0.75)
         || (cos(2*PI*(i-L0/2)/L0)<0.4 && sin(PI/4+2*PI*(k-D0/2)/D0)<0.4 && cos(2*PI*(i-L0/2)/L0)>0.2 && sin(PI/4+2*PI*(k-D0/2)/D0)>0.2 && k<L0/2)
         || (cos(PI*(i-L0/2)/L0)*cos(PI*(k-D0/2)/D0)<0.8 && cos(PI*(i-L0/2)/L0)*cos(PI*(k-D0/2)/D0)>0.75 && k>3*D0/5)))
           aux->celltype = GR;
      break;

#endif // MODEL_DUN

/****************************************************************************/
/************************** Templates of AVA model **************************/
/****************************************************************************/
#ifdef MODEL_AVA
    case CSP_SRC_CEIL:
    {
      /// CSP_SRC_CEIL: source of sand covering the ceiling
      if (j == 0) aux->celltype = IN;
      break;
    }

    case CSP_SRC_DISK_CEIL:
    {
      /// CSP_SRC_DISK(w,x,y): circular source of sand
      float w = p_temp->args[0];
      float x = p_temp->args[1];
      float y = p_temp->args[2];
      if ((j == 0) && ((i-x)*(i-x) + (k-y)*(k-y) <= w*w/4.0)) aux->celltype = IN;
      break;
    }
#endif // MODEL_AVA

/****************************************************************************/
/************************** Templates of CMB model **************************/
/****************************************************************************/
#ifdef MODEL_CMB
    case CSP_LAYER:
    {
      /// CSP_LAYER: fluid layer
      /// format: LAYER(h)
      int hh = (int)p_temp->args[0];
      float frac = fmodf(p_temp->args[0],1.0);
      if (!hh) hh = Hd2;
      if (j >= H0-hh) aux->celltype = MOINS;
      if ((j == H0-hh-1) && frac){ //h is not an integer
        if (drand48() < frac) aux->celltype = MOINS;
      }
      break;
    }
#endif // MODEL_CMB

/****************************************************************************/
/************************** Templates of ICB model **************************/
/****************************************************************************/
#ifdef MODEL_ICB
    case CSP_LAYER:
    {
      /// CSP_LAYER: fluid layer on top
      /// format: LAYER(h)
      int hh = (int)p_temp->args[0];
      float frac = fmodf(p_temp->args[0],1.0);
      if (!hh) hh = H0*2/3;
      if (j <= hh-1) aux->celltype = MOINS;
      if ((j == hh) && frac){ //h is not an integer
        if (drand48() < frac) aux->celltype = MOINS;
      }
      break;
    }
#endif // MODEL_ICB

/****************************************************************************/
/************************** Templates of RIV and D2G models **************************/
/****************************************************************************/
//#ifdef MODEL_RIV
#if defined(MODEL_RIV) || defined(MODEL_D2G)
    case CSP_LAYER:
    {
      /// CSP_LAYER: solid layer at the bottom
      /// format: LAYER(h)
      int hh = (int)p_temp->args[0];
      float frac = fmodf(p_temp->args[0],1.0);
      if (!hh) hh = Hd2;
      if (j >= H0-hh) aux->celltype = TERRE;
      if ((j == H0-1-hh) && frac){ //h is not an integer
        if (drand48() < frac) aux->celltype = TERRE;
      }
      break;
    }

    case CSP_SLOPE:
    {
      /// CSP_SLOPE: sloping ground
      /// format: SLOPE(r,h)
      float r = p_temp->args[0];
      int hh = (int)p_temp->args[1];
      if ((j >= H0-hh-((D0-1-k)*r)) && (aux->celltype != DUM)) aux->celltype = TERRE;
      break;
    }

    case CSP_CLOSED_BORDER:
    {
      /// CSP_CLOSED_BORDER: closed boundaries
      if ((i==0) || (k==0) || (i==L0-1) || (k==D0-1)) aux->celltype = DUM;
      break;
    }

    case CSP_OPEN_BORDER:
    {
      /// CSP_OPEN_BORDER: open boundaries above a given height
      /// format: OPEN_BORDER(h)
      int hh = (int)p_temp->args[0];
      if ((i==0) || (k==0) || (i==L0-1) || (k==D0-1)) aux->celltype = (j >= H0-hh) ? DUM : EAU;
      break;
    }

    case CSP_OPEN_BORDER1:
    {
      /// CSP_OPEN_BORDER1: rectangular outlet on the south border
      /// format: OPEN_BORDER1(h,w)
      int hh = (int)p_temp->args[0];
      int w = (int)p_temp->args[1];
      if ((i==0) || (k==0) || (i==L0-1)) aux->celltype = DUM;
      if (k==D0-1){ /// south border
        aux->celltype = EAU;
        if (j >= H0-hh) aux->celltype = DUM;
        if (w && fabs(i-Ld2)>=w/2) aux->celltype = DUM;
      }
      break;
    }

    case CSP_SRC:
    {
      /// CSP_SRC: source in the ceiling
      /// format SRC(x,y)
      int x = (int)p_temp->args[0];
      int y = (int)p_temp->args[1];
      if (!x) x = Ld2;
      if ((i==x) && (j==0) && (k==y)) aux->celltype = IN;
      break;
    }
#endif // MODEL_RIV

/****************************************************************************/
/************************** Templates of LIFE model **************************/
/****************************************************************************/
#ifdef MODEL_LIFE
    case CSP_RANDOM:
    {
      /// CSP_RANDOM: cells have probability <p> to be alive
      /// format RANDOM(p)
      float p = p_temp->args[0];
      if (drand48()<p) aux->celltype = ALIVE;
      break;
    }
#endif // MODEL_LIFE

    /*case CSP_CUSTOM: /// not supported
      break;*/
  } // end of switch

}

///////////////////////////////////////////////////////////////////////////////
/// Final state of a cell (rules applied over all space, whatever the templates)
///////////////////////////////////////////////////////////////////////////////
void init_cell_final(Cell *aux, int i, int j, int k)
{
  //int L0 = L-2*L_border;
  //int D0 = D-2*D_border;
  int H0 = H-2*H_border;

#if defined(MODEL_DUN) || defined(MODEL_AVA)
  /*** DUM cells in ground and ceiling ***/
  if (((j == H0-1) || (j==0)) && (aux->celltype != IN) && (aux->celltype != OUT)){
    aux->celltype   = DUM;
#ifdef CELL_COLOR
    aux->color = 0;
#endif
  }
#endif
}

void init_cell_from_template_array(Cell *aux, int i, int j, int k, int flag_add)
{
  int id_temp;

  /// templates are applied in reading order, thus leading to increasing priority
  //for (id_temp=nb_templates-1; id_temp>=0; id_temp--){
  for (id_temp=0; id_temp<nb_templates; id_temp++){
    if (!flag_add || csp_templates[id_temp].flag_add) init_cell_from_template(aux, i, j, k, id_temp);
  }
}

/// Init cellular space from template array.
/// When flag_add is set, select the array of added templates and modify the existing cellular space accordingly.
void init_csp_from_template_array(int flag_add)
{
  int i, j, k, x, y, z;
  Cell *aux;

  if (!flag_add){
    LogPrintf("init cellular space from template array\n");
    LogPrintf("nb_templates=%d\n", nb_templates);
  }
  else{
    LogPrintf("modify cellular space using the array of added templates\n");
    LogPrintf("nb_templates=%d\n", nb_templates);
    LogPrintf("nb_templates_add=%d\n", nb_templates_add);
  }

  /// initialization of the cellular space
  for(k=D_border, z=0; k < D-D_border; k++, z++){ /// north-south
    for(j=H_border, y=0; j < H-H_border; j++, y++){ /// vertical
      aux = TE + k*HL + j*L + L_border;
      for(i=L_border, x=0; i < L-L_border; i++, x++, aux++){ /// east-west
        //Assert(aux->celltype != BORD);

        if (!flag_add) init_cell_default(aux); /// default state

        init_cell_from_template_array(aux, x, y, z, flag_add);

        if (!flag_add) init_cell_final(aux, x, y, z);
      }
    }
  }
}

/// This function restores the type of a single cell in a template. When flag_rm is set,
/// the cell is attributed a type by default (according to the model and location).
void restore_template_cell(Cell *pt_csp, int i, int j, int k, unsigned char type, int flag_rm)
{
#ifdef MODEL_DUN
  if (flag_rm) type = ((j==H-1-H_border) || (j==H_border)) ? DUM : AIR;
  pt_csp[i+j*L+k*HL].celltype = type;
#endif // MODEL_DUN
}

/// The geometry of a solid template may be slightly altered after each rotation,
/// leading to a loss or gain of cells (as well as cell diffusion).
/// This function restores the original geometry with respect to the given orientation
/// (angle) of the cellular space.
/// When flag_rm is set, this function removes the solid templates.
/// The boolean argument flag_tmp signals a temporary reorientation of the cellular space.
void restore_template_geom(Cell *pt_csp, float angle, int flag_rm, int flag_tmp)
{
  int i, j, k, i0, k0;
  int id_temp, type_temp;
  char *name_temp;
  float *args_temp;
  static int start=1;
  int flag_warn;

  //if (start) LogPrintf("restore_template_geom: nb_templates = %d\n",nb_templates);

  /// use dimensions with no margin and no border
  int L0 = L-2*L_border;
  int D0 = D-2*D_border;
  //int H0 = H-2*H_border;

  //for (id_temp=nb_templates-1; id_temp>=0; id_temp--){
  for (id_temp=0; id_temp<nb_templates; id_temp++){

    type_temp = csp_templates[id_temp].type;
    name_temp = csp_templates[id_temp].name;
    args_temp = csp_templates[id_temp].args;
    flag_warn = 0;

    switch(type_temp){

#ifdef MODEL_DUN
      case CSP_SRC_UPWIND_CEIL:
      {
        /// source of sand in the ceiling parallel to the upwind boundaries
        if (flag_tmp) break; /// keep the template unchanged (when saving images)
        if (start) LogPrintf("restore the original geometry of template %s\n",name_temp);
        Assert(rot_map);
        for (k=1; k<D-1; k++){
          for (i=1; (i<L) && OutOfSpace(i,k); i++);
          i+=5;
          //LogPrintf("k=%d, i=%d, L=%d\n",k,i,L); //debug
          if ((i<L) && !OutOfSpace(i,k)) restore_template_cell(pt_csp, i, 1, k, IN, flag_rm);
        }
      }
      break;

      case CSP_SRC_DISK:
      case CSP_SRC_DISK_CEIL:
      {
        /// circular source (of sand) on top
        float w = args_temp[0]; //10.0;
        float x = args_temp[1]; //100;
        float y = args_temp[2]; //D>>1;
        j = (type_temp == CSP_SRC_DISK_CEIL) ? 1 : H-2;
        if (start) LogPrintf("restore the original geometry of template %s(%f,%f,%f)\n",name_temp,w,x,y);
        x += L_border - 0.5;
        y += D_border - 0.5;
        /// rotate and round the coordinates of the center
        Rotate_xz(x, y, angle);
        for (k=1; k<D-1; k++){
          for (i=1; i<L-1; i++){
            if (rot_map && OutOfSpace(i,k)) continue;
            /// restore circular source
            if ((i-x)*(i-x) + (k-y)*(k-y) <= w*w/4.0) restore_template_cell(pt_csp, i, j, k, IN, flag_rm);
          }
        }
      }
      break;

      case CSP_SRC_RECT:
      case CSP_SRC_RECT_CEIL:
      {
        /// rectangular source (of sand) on the ground
        float lx = args_temp[0]; //5.0;
        float ly = args_temp[1]; //D;
        float x = args_temp[2]; //100.0;
        float y = args_temp[3]; //D/2;
        float dx = lx*0.5;
        float dy = ly*0.5;
        j = (type_temp == CSP_SRC_RECT_CEIL) ? 1 : H-2;
        if (start) LogPrintf("restore the original geometry of template %s(%f,%f,%f,%f)\n",name_temp,lx,ly,x,y);
        x += L_border;
        y += D_border;
        for (k=1; k<D-1; k++){
          for (i=1; i<L-1; i++){
            if (rot_map && OutOfSpace(i,k)) continue;
            /// restore IN cells
            i0 = i;
            k0 = k;
            if (angle) Rotate_xz(i0, k0, -angle);
            if ((i0 >= x-dx) && (i0 < x+dx) && (k0 >= y-dy) && (k0 < y+dy)) restore_template_cell(pt_csp, i, j, k, IN, flag_rm);
          }
        }
      }
      break;

      case CSP_OUT_EAST:
      case CSP_OUT_NORTH:
      case CSP_OUT_SOUTH:
      {
        /// sink of sand along either the east, north or south border
        float w = args_temp[0]; //5.0;
        int cond = 0;
        if (start) LogPrintf("restore the original geometry of template %s(%f)\n",name_temp,w);
        j = H-2; /// ground
        for (k=1; k<D-1; k++){
          for (i=1; i<L-1; i++){
            if (rot_map && OutOfSpace(i,k)) continue;
            /// restore OUT cells
            i0 = i;
            k0 = k;
            if (angle) Rotate_xz(i0, k0, -angle);
            if (type_temp == CSP_OUT_EAST) cond = (i0 >= L-L_border-w);
            if (type_temp == CSP_OUT_NORTH) cond = (k0 <= D_border+w-1);
            if (type_temp == CSP_OUT_SOUTH) cond = (k0 >= D-D_border-w);
            if (cond) restore_template_cell(pt_csp, i, j, k, OUT, flag_rm);
          }
        }
      }
      break;

      case CSP_HBLOCK:
      case CSP_BLOCK:
      {
        /// 3-dimensionnal block
        i=0;
        int xmin = (int)args_temp[i++];
        int xmax = (int)args_temp[i++];
        int ymin = (int)args_temp[i++];
        int ymax = (int)args_temp[i++];
        int zmin = (type_temp == CSP_HBLOCK) ? 1 : args_temp[i++];
        int zmax = (type_temp == CSP_HBLOCK) ? 1 + args_temp[i++] : args_temp[i++];
        int type = (int)args_temp[i];

        if ((type!=DUM) && (type!=IN) && (type!=OUT)) break;

        if (start) LogPrintf("restore the original geometry of template BLOCK(%d,%d,%d,%d,%d,%d,%d)\n",xmin,xmax,ymin,ymax,zmin,zmax,type);

        xmin += L_border;
        xmax += L_border;
        ymin += D_border;
        ymax += D_border;
        zmin += H_border;
        zmax += H_border;

        Check_min_max(zmin, H_border, H-1-H_border);
        Check_min_max(zmax, H_border, H-1-H_border);

        for (k=1; k<D-1; k++){
          for (i=1; i<L-1; i++){
            if (rot_map && OutOfSpace(i,k)) continue;
            i0 = i;
            k0 = k;
            if (angle) Rotate_xz(i0, k0, -angle);
            /// restore block geometry
            if ((i0>=xmin) && (i0<xmax) && (k0>=ymin) && (k0<ymax)){
              for (j=H-1-zmin; j>H-1-zmax; j--) restore_template_cell(pt_csp, i, j, k, type, flag_rm);
            }
          }
        }
      }
      break;

      case CSP_CONE:
      case CSP_RCONE:
      {
        /// vertical cone (tip pointing upward)
        float hcone = (type_temp == CSP_CONE) ? args_temp[0] : args_temp[0]*(H-4);
        float lcone = (type_temp == CSP_CONE) ? args_temp[1] : args_temp[1]*Min(L0,D0);
        float x = (type_temp == CSP_CONE) ? args_temp[2] : args_temp[2]*L0;
        float y = (type_temp == CSP_CONE) ? args_temp[3] : args_temp[3]*D0;
        int type = (int)args_temp[4];

        if ((type!=DUM) && (type!=IN) && (type!=OUT)) break;

        if (start) LogPrintf("restore the original geometry of template CONE(%f,%f,%f,%f,%d)\n",hcone,lcone,x,y,type);

        x += L_border - 0.5;
        y += D_border - 0.5;

        /// rotate and round the coordinates of the center
        Rotate_xz(x, y, angle);

        int jmin = Max(H-3-hcone,1);
        for (j=H-3; j>jmin; j--){
          float rc = 0.5*lcone*(hcone-((H-3)-j))/hcone; /// radius
          for (k=1; k<D-1; k++){
            for (i=1; i<L-1; i++){
              if (rot_map && OutOfSpace(i,k)) continue;
              /// restore conical geometry
              if ((i-x)*(i-x) + (k-y)*(k-y) <= rc*rc) restore_template_cell(pt_csp, i, j, k, type, flag_rm);
            }
          }
        }
      }
      break;

      case CSP_RWALL:
      case CSP_STICKS:
      flag_warn = 1;
      break;
#else
      default: flag_warn = 1;
#endif // MODEL_DUN
    }
    if (start && flag_warn) WarnPrintf("WARNING: unable to restore template %s - not supported yet\n", name_temp);
  }

  start=0;
}


void templates_usage()
{
  init_template_types();

  printf("\nCSP TEMPLATES (%d)\n", nb_template_types);
  for(int i=0; i<nb_template_types; i++) printf("* %s\n", t_templates[i].desc);
}
